// cOlOr/C++
// e-mail: Franck.Barbier@FranckBarbier.com

#ifndef _Aggregation_h
#define _Aggregation_h

#ifdef _WIN32
#pragma warning (disable : 4786) // for debugging purposes with STL under Visual C++
#endif

#include <algorithm> // find generic function

#include "Association.h"

using namespace std;

template<typename T> class Composite : public Role<T> {
private:
    static list<T*> all_removed;
    static void add_in_all_removed(T*);
    static bool find_in_all_removed(T*);

private:
    set<T*, less<T*> > removed;

public:

    class CompositeException {
    };

public:
    Composite(T* t[] = NULL, int = 0, int = INT_MAX);
    Composite(const vector<void*>&, int = 0, int = INT_MAX);
    Composite(const Composite<T>&);
    virtual ~Composite();

public:
    virtual bool add(T*);
    virtual bool add(T&);
    virtual bool remove(T*);
    virtual bool remove(T&);

public:
    const Composite<T>& operator=(const Composite<T>&);
};

template<typename T> list<T*> Composite<T>::all_removed;

template<typename T> void Composite<T>::add_in_all_removed(T* t) {
    all_removed.push_back(t);
#ifdef _DEBUG
    list<T*>::iterator x;
    T* y;
    int z = 0;
    for (x = all_removed.begin(); x != all_removed.end(); x++) {
        y = *x;
        z++;
    }
#endif
}

template<typename T> bool Composite<T>::find_in_all_removed(T* t) {
#ifdef _DEBUG
    list<T*>::iterator x;
    T* y;
    int z = 0;
    for (x = all_removed.begin(); x != all_removed.end(); x++) {
        y = *x;
        z++;
    }
#endif
    typename list<T*>::iterator i = std::find(all_removed.begin(), all_removed.end(), t);
    if (i != all_removed.end()) return true;
    return false;
}

template<typename T> Composite<T>::Composite(T* t[], int i, int j) : Role<T>(t, i, j) {
}

template<typename T> Composite<T>::Composite(const vector<void*>& v, int i, int j) : Role<T>(v, i, j) {
}

template<typename T> Composite<T>::Composite(const Composite<T>& c) : Role<T>(c) {
}

template<typename T> Composite<T>::~Composite() {
    typename set<T*, less<T*> >::iterator i;
    typename list<T*>::iterator j;
    for (i = removed.begin(); i != removed.end(); i++) {
        j = std::find(all_removed.begin(), all_removed.end(), *i);
        if (j != all_removed.end()) all_removed.erase(j);
        else {
            // in C++11 destructors default to noexcept
            // if (all_removed.size()) throw CompositeException();
        }
    }
}

template<typename T> bool Composite<T>::add(T* t) {
    if (!Composite<T>::find_in_all_removed(t)) {
        if (Role<T>::add(t)) return true;
    }
    return false;
}

template<typename T> bool Composite<T>::add(T& t) {
    return add(&t);
}

template<typename T> bool Composite<T>::remove(T* t) {
    if (Role<T>::remove(t)) {
        Composite<T>::add_in_all_removed(t);
        if (!removed.insert(t).second) throw CompositeException();
        return true;
    }
    return false;
}

template<typename T> bool Composite<T>::remove(T& t) {
    return remove(&t);
}

template<typename T> const Composite<T>& Composite<T>::operator=(const Composite<T>& c) {
    this->Role<T>::operator=(c);
    if (this != &c) {
        typename set<T*, less<T*> >::iterator i;
        typename list<T*>::iterator j;
        if (!removed.empty()) {
            for (i = removed.begin(); i != removed.end(); i++) {
                j = std::find(all_removed.begin(), all_removed.end(), *i);
                if (j != all_removed.end()) all_removed.erase(j);
                else throw CompositeException();
            }
        }
        removed = c.removed;
    }
    return *this;
}

template<typename T> class Compound : public Role<T> {
private:
    static list<T*> all_removed;
    static void add_in_all_removed(T*);
    static bool find_in_all_removed(T*);

private:
    set<T*, less<T*> > removed;

public:

    class CompoundException {
    };

public:
    Compound(T* t[] = NULL, int = 0, int = INT_MAX);
    Compound(const vector<void*>& v, int = 0, int = INT_MAX);
    Compound(const Compound<T>&);
    virtual ~Compound();

public:
    virtual bool add(T*);
    virtual bool add(T&);
    virtual bool remove(T*);
    virtual bool remove(T&);

public:
    const Compound<T>& operator=(const Compound<T>&);
};

template<typename T> list<T*> Compound<T>::all_removed;

template<typename T> void Compound<T>::add_in_all_removed(T* t) {
    all_removed.push_back(t);
#ifdef _DEBUG
    list<T*>::iterator x;
    T* y;
    int z = 0;
    for (x = all_removed.begin(); x != all_removed.end(); x++) {
        y = *x;
        z++;
    }
#endif
}

template<typename T> bool Compound<T>::find_in_all_removed(T* t) {
#ifdef _DEBUG
    list<T*>::iterator x;
    T* y;
    int z = 0;
    for (x = all_removed.begin(); x != all_removed.end(); x++) {
        y = *x;
        z++;
    }
#endif
    typename list<T*>::iterator i = std::find(all_removed.begin(), all_removed.end(), t);
    if (i != all_removed.end()) return true;
    return false;
}

template<typename T> Compound<T>::Compound(T* t[], int i, int j) : Role<T>(t, i, j) {
}

template<typename T> Compound<T>::Compound(const vector<void*>& v, int i, int j) : Role<T>(v, i, j) {
}

template<typename T> Compound<T>::Compound(const Compound<T>& c) : Role<T>(c) {
}

template<typename T> Compound<T>::~Compound() {
    typename set<T*, less<T*> >::iterator i;
    typename list<T*>::iterator j;
    for (i = removed.begin(); i != removed.end(); i++) {
        j = std::find(all_removed.begin(), all_removed.end(), *i);
        if (j != all_removed.end()) all_removed.erase(j);
        else {
            // in C++11 destructors default to noexcept
            // if (all_removed.size()) throw CompoundException();
        }
    }
}

template<typename T> bool Compound<T>::add(T* t) {
    if (!Compound<T>::find_in_all_removed(t)) {
        if (Role<T>::add(t)) return true;
    }
    return false;
}

template<typename T> bool Compound<T>::add(T& t) {
    return add(&t);
}

template<typename T> bool Compound<T>::remove(T* t) {
    if (Role<T>::remove(t)) {
        Compound<T>::add_in_all_removed(t);
        if (!removed.insert(t).second) throw CompoundException();
        return true;
    }
    return false;
}

template<typename T> bool Compound<T>::remove(T& t) {
    return remove(&t);
}

template<typename T> const Compound<T>& Compound<T>::operator=(const Compound<T>& c) {
    this->Role<T>::operator=(c);
    if (this != &c) {
        typename set<T*, less<T*> >::iterator i;
        typename list<T*>::iterator j;
        if (!removed.empty()) {
            for (i = removed.begin(); i != removed.end(); i++) {
                j = std::find(all_removed.begin(), all_removed.end(), *i);
                if (j != all_removed.end()) all_removed.erase(j);
                else throw CompoundException();
            }
        }
        removed = c.removed;
    }
    return *this;
}

template<typename Whole, typename Part> class Aggregation;

template<typename Whole, typename Part> class _Aggregation {
    friend class Aggregation<Whole, Part>;

private:
    map<Whole*, Composite<Part>, less<Whole*> > all_parts;
    map<Part*, Compound<Whole>, less<Part*> > all_wholes;
    int min_Whole;
    int max_Whole;
    int min_Part;
    int max_Part;

private:
    list<Compound<Whole>* > garbage_Whole;
    list<Composite<Part>* > garbage_Part;

private:

    class _AggregationException {
    };

public: // for STL map
    _Aggregation(int = 0, int = INT_MAX, int = 0, int = INT_MAX);
    ~_Aggregation();

private:
    bool add(Whole*, Part*);
    bool add(Whole&, Part&);
    bool add(Part*, Whole*);
    bool add(Part&, Whole&);
    bool add(Whole*, Composite<Part>&);
    bool add(Whole&, Composite<Part>&);
    bool add(Part*, Compound<Whole>&);
    bool add(Part&, Compound<Whole>&);
    bool multiplicity_conformity();
    int multiplicity(Whole*);
    int multiplicity(Whole&);
    int multiplicity(Part*);
    int multiplicity(Part&);
    bool remove(Whole*);
    bool remove(Whole&);
    bool remove(Part*);
    bool remove(Part&);
    bool remove(Whole*, Part*);
    bool remove(Whole&, Part&);
    bool remove(Part*, Whole*);
    bool remove(Part&, Whole&);
    Composite<Part> scan(Whole*);
    Composite<Part> scan(Whole&);
    Composite<Part> scan(Compound<Whole>&);
    Compound<Whole> scan(Part*);
    Compound<Whole> scan(Part&);
    Compound<Whole> scan(Composite<Part>&);
    bool test(Whole*);
    bool test(Whole&);
    bool test(Part*);
    bool test(Part&);
    bool test(Whole*, Part*);
    bool test(Whole&, Part&);
    bool test(Part*, Whole*);
    bool test(Part&, Whole&);
};

template<typename Whole, typename Part> _Aggregation<Whole, Part>::_Aggregation(int i, int j, int k, int l) {
    if (i >= 0 && k >= 0 && j > 0 && l > 0 && i <= j && k <= l) {
        min_Whole = i;
        max_Whole = j;
        min_Part = k;
        max_Part = l;
    } else throw _AggregationException();
}

template<typename Whole, typename Part> _Aggregation<Whole, Part>::~_Aggregation() {
    typename list<Compound<Whole>* >::iterator i;
    typename list<Composite<Part>* >::iterator j;
    for (i = garbage_Whole.begin(); i != garbage_Whole.end(); i++) delete (*i);
    for (j = garbage_Part.begin(); j != garbage_Part.end(); j++) delete (*j);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::add(Whole* whole, Part* part) {
    if (!all_parts.count(whole)) {
        // whole does not exist
        if (!all_wholes.count(part)) {
            // whole does not exist and part does not exist
            Compound<Whole>* wholes = new Compound<Whole>();
            if (!wholes->add(whole)) {
                delete wholes;
                return false;
            }
            garbage_Whole.push_back(wholes);
            Composite<Part>* parts = new Composite<Part>();
            if (!parts->add(part)) {
                delete parts;
                return false;
            }
            garbage_Part.push_back(parts);
            all_wholes[part] = *wholes;
            all_parts[whole] = *parts;
        } else {
            // whole does not exist but part already exists
            assert(all_wholes[part].multiplicity() <= max_Whole && !all_wholes[part].test(whole));
            if (all_wholes[part].multiplicity() == max_Whole) return false;
            Composite<Part>* parts = new Composite<Part>();
            if (!parts->add(part)) {
                delete parts;
                return false;
            }
            garbage_Part.push_back(parts);
            if (!all_wholes[part].add(whole)) return false;
            all_parts[whole] = *parts;
        }
    } else {
        // whole already exists
        assert(all_parts[whole].multiplicity() <= max_Part);
        if (all_parts[whole].multiplicity() == max_Part || all_parts[whole].test(part)) return false;
        if (!all_wholes.count(part)) {
            // whole already exists but part does not exist
            Compound<Whole>* wholes = new Compound<Whole>();
            if (!wholes->add(whole)) {
                delete wholes;
                return false;
            }
            garbage_Whole.push_back(wholes);
            if (!all_parts[whole].add(part)) return false;
            all_wholes[part] = *wholes;
        } else {
            // whole already exists and part already exists with no link between them
            assert(all_wholes[part].multiplicity() <= max_Whole && !all_wholes[part].test(whole));
            if (all_wholes[part].multiplicity() == max_Whole) return false;
            Compound<Whole>* wholes = new Compound<Whole>();
            if (!wholes->add(whole)) {
                delete wholes;
                return false;
            }
            delete wholes;
            Composite<Part>* parts = new Composite<Part>();
            if (!parts->add(part)) {
                delete parts;
                return false;
            }
            delete parts;
            if (!all_parts[whole].add(part)) throw _AggregationException();
            if (!all_wholes[part].add(whole)) throw _AggregationException();
        }
    }
    return true;
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::add(Whole& whole, Part& part) {
    return add(&whole, &part);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::add(Part* part, Whole* whole) {
    return add(whole, part);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::add(Part& part, Whole& whole) {
    return add(&whole, &part);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::add(Whole* whole, Composite<Part>& parts) {
    bool result = true;
    Part* part = parts.head();
    while (part) {
        if (!add(whole, part)) result = false;
        part = parts.next();
    }
    return result;
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::add(Whole& whole, Composite<Part>& parts) {
    return add(&whole, parts);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::add(Part* part, Compound<Whole>& wholes) {
    bool result = true;
    Whole* whole = wholes.head();
    while (whole) {
        if (!add(part, whole)) result = false;
        whole = wholes.next();
    }
    return result;
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::add(Part& part, Compound<Whole>& wholes) {
    return add(&part, wholes);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::multiplicity_conformity() {
    typename map<Whole*, Composite<Part>, less<Whole*> >::iterator i;
    for (i = all_parts.begin(); i != all_parts.end(); i++) {
        assert((*i).second.multiplicity() <= max_Part);
        if ((*i).second.multiplicity() < min_Part) return false;
    }
    typename map<Part*, Compound<Whole>, less<Part*> >::iterator j;
    for (j = all_wholes.begin(); j != all_wholes.end(); j++) {
        assert((*j).second.multiplicity() <= max_Whole);
        if ((*j).second.multiplicity() < min_Whole) return false;
    }
    return true;
}

template<typename Whole, typename Part> int _Aggregation<Whole, Part>::multiplicity(Whole* whole) {
    if (all_parts.count(whole)) return all_parts[whole].multiplicity();
    return -1;
}

template<typename Whole, typename Part> int _Aggregation<Whole, Part>::multiplicity(Whole& whole) {
    return multiplicity(&whole);
}

template<typename Whole, typename Part> int _Aggregation<Whole, Part>::multiplicity(Part* part) {
    if (all_wholes.count(part)) return all_wholes[part].multiplicity();
    return -1;
}

template<typename Whole, typename Part> int _Aggregation<Whole, Part>::multiplicity(Part& part) {
    return multiplicity(&part);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::remove(Whole* whole) {
    if (all_parts.count(whole)) {
        Part* part = all_parts[whole].head();
        while (part) {
            if (all_wholes.count(part)) {
                assert(all_wholes[part].test(whole));
                if (multiplicity_conformity() && all_wholes[part].multiplicity() == min_Whole) return false;
            } else throw _AggregationException();
            part = all_parts[whole].next();
        }
        part = all_parts[whole].head();
        while (part) {
            if (all_wholes.count(part)) {
                if (!all_wholes[part].remove(whole)) throw _AggregationException();
                if (!all_wholes[part].multiplicity()) {
                    if (!all_wholes.erase(part)) throw _AggregationException();
                }
            } else throw _AggregationException();
            part = all_parts[whole].next();
        }
        if (!all_parts.erase(whole)) throw _AggregationException();
        return true;
    }
    return false;
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::remove(Whole& whole) {
    return remove(&whole);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::remove(Part* part) {
    if (all_wholes.count(part)) {
        Whole* whole = all_wholes[part].head();
        while (whole) {
            if (all_parts.count(whole)) {
                assert(all_parts[whole].test(part));
                if (multiplicity_conformity() && all_parts[whole].multiplicity() == min_Part) return false;
            } else throw _AggregationException();
            whole = all_wholes[part].next();
        }
        whole = all_wholes[part].head();
        while (whole) {
            if (all_parts.count(whole)) {
                if (!all_parts[whole].remove(part)) throw _AggregationException();
                if (!all_parts[whole].multiplicity()) {
                    if (!all_parts.erase(whole)) throw _AggregationException();
                }
            } else throw _AggregationException();
            whole = all_wholes[part].next();
        }
        if (!all_wholes.erase(part)) throw _AggregationException();
        return true;
    }
    return false;
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::remove(Part& part) {
    return remove(&part);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::remove(Whole* whole, Part* part) {
    if (!all_parts.count(whole)) return false;
    else {
        if (!all_wholes.count(part)) return false;
        else {
            if (multiplicity_conformity() &&
                    (all_parts[whole].multiplicity() == min_Part ||
                    all_wholes[part].multiplicity() == min_Whole)) return false;
            if (!all_parts[whole].test(part)) {
                assert(!all_wholes[part].test(whole));
                return false;
            }
        }
    }
    if (!all_parts[whole].remove(part)) throw _AggregationException();
    if (!all_wholes[part].remove(whole)) throw _AggregationException();
    if (!all_parts[whole].multiplicity())
        if (!all_parts.erase(whole)) throw _AggregationException();
    if (!all_wholes[part].multiplicity())
        if (!all_wholes.erase(part)) throw _AggregationException();
    return true;
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::remove(Whole& whole, Part& part) {
    return remove(&whole, &part);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::remove(Part* part, Whole* whole) {
    return remove(whole, part);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::remove(Part& part, Whole& whole) {
    return remove(&part, &whole);
}

template<typename Whole, typename Part> Composite<Part> _Aggregation<Whole, Part>::scan(Whole* whole) {
    if (all_parts.count(whole)) return all_parts[whole];
    return Composite<Part>();
}

template<typename Whole, typename Part> Composite<Part> _Aggregation<Whole, Part>::scan(Whole& whole) {
    return scan(&whole);
}

template<typename Whole, typename Part> Composite<Part> _Aggregation<Whole, Part>::scan(Compound<Whole>& wholes) {
    Composite<Part> result;
    Whole* whole = wholes.head();
    while (whole) {
        Composite<Part> parts = scan(whole);
        (void) set_union(result.begin(), result.end(), parts.begin(), parts.end(), std::inserter(result, result.end()));
        whole = wholes.next();
    }
    return result;
}

template<typename Whole, typename Part> Compound<Whole> _Aggregation<Whole, Part>::scan(Part* part) {
    if (all_wholes.count(part)) return all_wholes[part];
    return Compound<Whole>();
}

template<typename Whole, typename Part> Compound<Whole> _Aggregation<Whole, Part>::scan(Part& part) {
    return scan(&part);
}

template<typename Whole, typename Part> Compound<Whole> _Aggregation<Whole, Part>::scan(Composite<Part>& parts) {
    Compound<Whole> result;
    Part* part = parts.head();
    while (part) {
        Compound<Whole> wholes = scan(part);
        (void) ::set_union(result.begin(), result.end(), wholes.begin(), wholes.end(), std::inserter(result, result.end()));
        part = parts.next();
    }
    return result;
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::test(Whole* whole) {
    if (all_parts.count(whole)) return true;
    return false;
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::test(Whole& whole) {
    return test(&whole);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::test(Part* part) {
    if (all_wholes.count(part)) return true;
    return false;
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::test(Part& part) {
    return test(&part);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::test(Whole* whole, Part* part) {
    if (all_parts.count(whole) && all_parts[whole].test(part)) return true;
    return false;
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::test(Whole& whole, Part& part) {
    return test(&whole, &part);
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::test(Part* part, Whole* whole) {
    if (all_wholes.count(part) && all_wholes[part].test(whole)) return true;
    return false;
}

template<typename Whole, typename Part> bool _Aggregation<Whole, Part>::test(Part& part, Whole& whole) {
    return test(&part, &whole);
}

template<typename Whole, typename Part> class Aggregation {
private:
    static map<string, _Aggregation<Whole, Part>, less<string> > _aggregation;
    string _name;

public:

    class AggregationException {
    };

public:
    Aggregation(int = 0, int = INT_MAX, int = 0, int = INT_MAX, string = "");
    ~Aggregation();

public:
    bool add(Whole*, Part*);
    bool add(Whole&, Part&);
    bool add(Part*, Whole*);
    bool add(Part&, Whole&);
    bool add(Whole*, Composite<Part>&);
    bool add(Whole&, Composite<Part>&);
    bool add(Part*, Compound<Whole>&);
    bool add(Part&, Compound<Whole>&);
    bool multiplicity_conformity();
    int multiplicity(Whole*);
    int multiplicity(Whole&);
    int multiplicity(Part*);
    int multiplicity(Part&);
    bool remove(Whole*);
    bool remove(Whole&);
    bool remove(Part*);
    bool remove(Part&);
    bool remove(Whole*, Part*);
    bool remove(Whole&, Part&);
    bool remove(Part*, Whole*);
    bool remove(Part&, Whole&);
    Composite<Part> scan(Whole*);
    Composite<Part> scan(Whole&);
    Composite<Part> scan(Compound<Whole>&);
    Compound<Whole> scan(Part*);
    Compound<Whole> scan(Part&);
    Compound<Whole> scan(Composite<Part>&);
    bool test(Whole*);
    bool test(Whole&);
    bool test(Part*);
    bool test(Part&);
    bool test(Whole*, Part*);
    bool test(Whole&, Part&);
    bool test(Part*, Whole*);
    bool test(Part&, Whole&);
};

template<typename Whole, typename Part> map<string, _Aggregation<Whole, Part>, less<string> > Aggregation<Whole, Part>::_aggregation;

template<typename Whole, typename Part> Aggregation<Whole, Part>::Aggregation(int i, int j, int k, int l, string name) {
    _name = name;
    if (_aggregation.find(_name) == _aggregation.end()) {
        try {
            _aggregation[_name] = _Aggregation<Whole, Part>(i, j, k, l);
        } catch (typename _Aggregation<Whole, Part>::_AggregationException) {
            throw AggregationException();
        }
    }
}

template<typename Whole, typename Part> Aggregation<Whole, Part>::~Aggregation() {
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::add(Whole* whole, Part* part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].add(whole, part);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::add(Whole& whole, Part& part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].add(whole, part);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::add(Part* part, Whole* whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].add(part, whole);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::add(Part& part, Whole& whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].add(part, whole);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::add(Whole* whole, Composite<Part>& parts) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].add(whole, parts);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::add(Whole& whole, Composite<Part>& parts) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].add(whole, parts);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::add(Part* part, Compound<Whole>& wholes) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].add(part, wholes);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::add(Part& part, Compound<Whole>& wholes) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].add(part, wholes);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::multiplicity_conformity() {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].multiplicity_conformity();
}

template<typename Whole, typename Part> int Aggregation<Whole, Part>::multiplicity(Whole* whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].multiplicity(whole);
}

template<typename Whole, typename Part> int Aggregation<Whole, Part>::multiplicity(Whole& whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].multiplicity(whole);
}

template<typename Whole, typename Part> int Aggregation<Whole, Part>::multiplicity(Part* part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].multiplicity(part);
}

template<typename Whole, typename Part> int Aggregation<Whole, Part>::multiplicity(Part& part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].multiplicity(part);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::remove(Whole* whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].remove(whole);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::remove(Whole& whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].remove(whole);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::remove(Part* part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].remove(part);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::remove(Part& part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].remove(part);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::remove(Whole* whole, Part* part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].remove(whole, part);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::remove(Whole& whole, Part& part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].remove(whole, part);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::remove(Part* part, Whole* whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].remove(part, whole);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::remove(Part& part, Whole& whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].remove(part, whole);
}

template<typename Whole, typename Part> Composite<Part> Aggregation<Whole, Part>::scan(Whole* whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].scan(whole);
}

template<typename Whole, typename Part> Composite<Part> Aggregation<Whole, Part>::scan(Whole& whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].scan(whole);
}

template<typename Whole, typename Part> Composite<Part> Aggregation<Whole, Part>::scan(Compound<Whole>& wholes) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].scan(wholes);
}

template<typename Whole, typename Part> Compound<Whole> Aggregation<Whole, Part>::scan(Part* part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].scan(part);
}

template<typename Whole, typename Part> Compound<Whole> Aggregation<Whole, Part>::scan(Part& part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].scan(part);
}

template<typename Whole, typename Part> Compound<Whole> Aggregation<Whole, Part>::scan(Composite<Part>& parts) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].scan(parts);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::test(Whole* whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].test(whole);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::test(Whole& whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].test(whole);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::test(Part* part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].test(part);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::test(Part& part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].test(part);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::test(Whole* whole, Part* part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].test(whole, part);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::test(Whole& whole, Part& part) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].test(whole, part);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::test(Part* part, Whole* whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].test(part, whole);
}

template<typename Whole, typename Part> bool Aggregation<Whole, Part>::test(Part& part, Whole& whole) {
    assert(_aggregation.find(_name) != _aggregation.end());
    return _aggregation[_name].test(part, whole);
}

#endif
